import { createMachine, assign, spawn, actions, send, interpret } from 'xstate';
import { createRunner, OBJ, ARR, FUNC } from 'ploson';
import * as R from 'ramda';

import { Rule } from './rule-dm';
import { ruleMachine, runPlosonOrReject } from './rule';
import { configuredCollectionDiff, idUniquesInFirst } from './utils';

const { pure, stop } = actions;

const groupByHasCallback = R.groupBy(({ callback }) => (callback ? 'updatedSoft' : 'updatedHard'));
const populateCallbacks = R.map(({ trigger, ploson, ...item }) =>
  runPlosonOrReject({
    ploson,
    code: trigger,
    createReturn: (callback) => ({ ...item, trigger, ploson, callback }),
  }),
);

const getTriggerCallbackActor = (callback, targets) => (sendCb) =>
  callback((payload, dynamicTargets) => {
    // console.log('payload:', payload);
    return sendCb({ type: 'FIRE_RULES', targets: dynamicTargets ?? targets, payload });
  });
const spawnTriggerActors = R.map(({ id, targets, callback, ...item }) => ({
  id,
  targets,
  callback,
  ...item,
  ...(callback ? { ref: spawn(getTriggerCallbackActor(callback, targets), id) } : OBJ),
}));

const context = {
  rules: [
    // { ploson, ref, ...rule }
  ],
  triggers: [
    // { id, trigger, targets, ploson, callback, ref }
  ],
};

const machineConf = {
  predictableActionArguments: true,
  id: 'rule-manager',
  initial: 'initRules',
  context,
  states: {
    retryRules: { after: { RETRY_DELAY: { target: 'initRules' } } },
    initRules: {
      invoke: {
        src: 'ruleFetcher',
        onDone: 'prepareRules',
        onError: { actions: 'logError', target: 'retryRules' },
      },
    },
    prepareRules: {
      invoke: {
        id: 'processInitRules',
        src: 'processRules',
        onDone: { actions: 'addRules', target: 'initTriggers' },
        onError: { actions: 'logError', target: 'initTriggers' },
      },
    },
    retryTriggers: { after: { RETRY_DELAY: { target: 'initTriggers' } } },
    initTriggers: {
      invoke: {
        src: 'triggerFetcher',
        onDone: { target: 'prepareTriggers' },
        onError: { actions: 'logError', target: 'retryTriggers' },
      },
    },
    prepareTriggers: {
      invoke: {
        id: 'processInitTriggers',
        src: 'processTriggers',
        onDone: { actions: ['addTriggers', 'spawnUpdater'], target: 'running' },
        onError: { actions: 'logError', target: 'running' },
      },
    },
    running: {
      initial: 'idle',
      on: {
        FIRE_RULES: { actions: 'fireRules' },
      },
      states: {
        idle: {
          on: {
            UPDATE_RULES: { target: 'updatingRules' },
            UPDATE_TRIGGERS: { target: 'updatingTriggers' },
          },
        },
        updatingRules: {
          invoke: {
            id: 'processRuleUpdates',
            src: 'processRules',
            onDone: {
              target: 'idle',
              actions: ['killActors', 'removeRules', 'addRules', 'updateRules'],
            },
            onError: { target: 'idle', actions: 'logError' },
          },
        },
        updatingTriggers: {
          invoke: {
            id: 'processTriggerUpdates',
            src: 'processTriggers',
            onDone: {
              target: 'idle',
              actions: ['killActors', 'removeTriggers', 'addTriggers'],
            },
            onError: { target: 'idle', actions: 'logError' },
          },
        },
      },
    },
  },
};

const getOptions = ({
  ruleFetcher = () => Promise.resolve(ARR),
  triggerFetcher = () => Promise.resolve(ARR),
  staticEnv = OBJ,
  staticEnvTriggers = staticEnv,
  updater = FUNC,
}) => {
  const rulesDiff = configuredCollectionDiff({
    ignoreKeys: ['ref', 'ploson', 'active'],
    skipIf: ({ active }) => !active,
    transformAdded: (item) => {
      const { value, error, warning } = Rule.validate(item);
      if (error || warning) console.warn(error || warning);
      const ploson = createRunner({ staticEnv });
      return { ...value, ploson };
    },
    transformUpdated: (item, { ploson, ref }) => ({ ...item, ploson, ref }),
  });

  const triggersDiff = configuredCollectionDiff({
    ignoreKeys: ['ploson', 'callback', 'ref', 'active'],
    skipIf: ({ active }) => !active,
    transformAdded: ({ targets = ARR, ...item }) => {
      const ploson = createRunner({ staticEnv: staticEnvTriggers });
      return { ...item, targets, ploson };
    },
    transformUpdated: (item, { ploson, callback, ref }, patch) => ({
      ...item,
      ploson,
      ref,
      ...(R.any(R.propEq('path', '/trigger'))(patch) ? OBJ : { callback }),
    }),
  });

  return {
    delays: { RETRY_DELAY: 30_000 },

    services: {
      ruleFetcher,
      triggerFetcher,
      processRules: async ({ rules }, { data = ARR }) => {
        const { added, updated, removed } = rulesDiff(rules)(data);
        await Promise.all(
          R.map(({ onLoad, ploson }) => {
            if (!onLoad) return Promise.resolve();
            return runPlosonOrReject({ ploson, code: onLoad });
          })([...added, ...updated]),
        );
        // console.log('Rules processed (onLoad) and ready to be run:', added, removed, updated);
        return { added, toRemove: removed, toUpdate: updated };
      },
      processTriggers: async ({ triggers }, { data = ARR }) => {
        const { added, updated, removed } = triggersDiff(triggers)(data);
        const { updatedHard = ARR, updatedSoft = ARR } = groupByHasCallback(updated);
        const toAdd = await Promise.all(populateCallbacks([...added, ...updatedHard]));
        return { added: toAdd, toRemove: [...removed, ...updatedHard], toUpdate: updatedSoft };
      },
    },

    actions: {
      logError: (ctx, { data }) => console.error(data),
      killActors: pure((ctx, { data: { toRemove = ARR, toRestart = ARR } = OBJ }) => {
        const events = R.into(
          [],
          R.pipe(
            R.filter(({ ref }) => Boolean(ref)),
            R.map(({ ref }) => stop(ref)),
          ),
        )([...toRemove, ...toRestart]);
        // console.log('killActors events:', events);
        return events;
      }),
      removeTriggers: assign({
        triggers: ({ triggers }, { data: { toRemove = ARR } = OBJ }) =>
          idUniquesInFirst(triggers, toRemove),
        // TODO: Destructors??
      }),
      addTriggers: assign({
        triggers: ({ triggers }, { data: { added = ARR } = OBJ }) => {
          console.info('Spawning triggers:', R.pluck('id', added));
          const addedTriggers = spawnTriggerActors(added);
          return [...triggers, ...addedTriggers];
        },
      }),
      updateTriggers: assign({
        triggers: ({ triggers }, { data: { toUpdate = ARR } = OBJ }) => {
          // R.over(R.lensIndex(R.findIndex(R.propEq('id', id))), R.assoc('targets'))(triggers);
          return R.map((item) => {
            const updated = R.find(R.propEq('id', item.id))(toUpdate);
            return updated ? R.assoc('targets', updated.targets ?? [], item) : item;
          })(triggers);
        },
      }),
      removeRules: assign({
        rules: ({ rules }, { data: { toRemove = ARR } = OBJ }) => idUniquesInFirst(rules, toRemove),
      }),
      addRules: assign({
        rules: ({ rules }, { data: { added = ARR } = OBJ }) => {
          const addedRules = R.map(
            ({
              id,
              active,
              ttl,
              resetCondition,
              resetAction,
              cooldown,
              condition,
              action,
              ploson,
              ...rest
            }) => {
              const hysteresis = Boolean(resetCondition);
              const ctx = {
                ttl,
                hysteresis,
                condition,
                action,
                ploson,
                resetCondition,
                resetAction,
                cooldown,
              };
              // console.log('Spawning Rule with id:', id);
              // console.log(active, ctx);
              const spawnedIfActive = active
                ? { ref: spawn(ruleMachine.withContext(ctx), id) }
                : OBJ;
              return { id, active, ttl, ploson, ...rest, ...spawnedIfActive };
            },
          )(added);
          return [...rules, ...addedRules];
        },
      }),
      updateRules: pure((ctx, { data: { toUpdate = ARR } = OBJ }) => {
        const actionMatrix = R.chain(
          ({ active, ttl, resetCondition, resetAction, cooldown, condition, action, ref }) => {
            const hysteresis = Boolean(resetCondition);
            const rule = {
              ttl,
              hysteresis,
              condition,
              action,
              resetCondition,
              resetAction,
              cooldown,
            };
            const updateEvent = send({ type: 'UPDATE', rule }, { to: ref });
            return !active ? [send('INACTIVATE', { to: ref }), updateEvent] : [updateEvent];
          },
        )(toUpdate);
        return actionMatrix;
      }),
      fireRules: pure(({ rules }, { targets, payload }) => {
        // console.log('fireRules...:', targets, payload);
        const events = R.reduce(
          (acc, { id, active, ref }) =>
            active && ref && R.includes(id, targets)
              ? [...acc, send({ type: 'FIRE', payload }, { to: ref })]
              : acc,
          [],
        )(rules);
        // console.log('fireRules:', events);
        return events;
      }),
      spawnUpdater: assign({ updaterRef: () => spawn(updater, 'rule-trigger-updater') }),
    },
  };
};

export const createRuleEngineMachine = (config) => {
  return createMachine(machineConf, getOptions(config));
};

export const createRuleEngine = (...args) => interpret(createRuleEngineMachine(...args));

/**
- active->inactive for triggers should clean up actor
*/
